"""
Модуль, содержащий репозиторий локального файлового хранилища.
"""

import asyncio
import os
from collections.abc import AsyncGenerator, Awaitable, Callable
from contextlib import asynccontextmanager
from logging import getLogger
from pathlib import Path
from typing import Self, cast

from aiofiles import open
from aiofiles import os as aos

from .reader import AiofilesStreamReader, StreamReaderProtocol, StreamReadProtocol
from .schemas import LocalStorageParamsSchema


class LocalStorageRepository:
    """
    Репозиторий локального файлового хранилища.
    """

    def __init__(self: Self, params: LocalStorageParamsSchema) -> None:
        self.work_dir = Path(params.path)
        if not os.path.exists(self.work_dir):
            os.makedirs(self.work_dir)
        self.logger = getLogger(__name__)

    async def exists(self: Self, path: str | Path) -> bool:
        """
        Проверяем существует ли файл.
        """
        path = self.work_dir / path
        return await aos.path.exists(path)

    async def listdir(self: Self, path: str | Path) -> list[str]:
        """
        Получаем список файлов и директорий в заданной директории.
        """
        paths: list[str] = []
        for item in await aos.listdir(self.work_dir / path):
            item_path = str(Path(path) / item)
            if await aos.path.isdir(self.work_dir / item_path):
                item_path += '/'
            paths.append(item_path)
        return paths

    async def is_file(self: Self, path: str | Path) -> bool:
        """
        Проверяем находится ли файл по пути.
        """
        path = self.work_dir / path
        return await aos.path.isfile(path)

    async def is_dir(self: Self, path: str | Path) -> bool:
        """
        Проверяем находится ли директория по пути.
        """
        path = self.work_dir / path
        return await aos.path.isdir(path)

    async def read(self: Self, path: str | Path) -> bytes:
        """
        Читаем содержимое файла.
        """
        path = self.work_dir / path
        async with open(path, 'rb') as f:
            return await f.read()

    @asynccontextmanager
    async def stream_read(self: Self, path: str | Path) -> AsyncGenerator[StreamReaderProtocol, None]:
        """
        Читаем содержимое файла в потоковом режиме.
        """
        path = self.work_dir / path
        async with open(path, 'rb') as f:
            yield AiofilesStreamReader(f)

    async def write(self: Self, path: str | Path, content: str | bytes) -> None:
        """
        Создаем файл или переписываем существующий.
        """
        path = self.work_dir / path
        await aos.makedirs(path.parent, exist_ok=True)
        async with open(path, mode='wb') as f:
            await f.write(content.encode('utf-8') if isinstance(content, str) else content)

    async def stream_write(
        self: Self,
        path: str | Path,
        stream: StreamReadProtocol,
        length: int = -1,
        part_size: int = 0,
    ) -> None:
        """
        Создаем файл или переписываем существующий в потоковом режиме.
        """
        if length != -1:
            self.logger.warning('Параметр length не используется для LocalStorage.')
        part_size = part_size or 1024
        path = self.work_dir / path
        is_co_function = asyncio.iscoroutinefunction(stream.read)
        async with open(path, 'wb') as f:
            while chunk := (
                await cast(Callable[[int], Awaitable[bytes]], stream.read)(part_size)
                if is_co_function
                else cast(Callable[[int], bytes], stream.read)(part_size)
            ):
                await f.write(chunk)

    async def delete(self: Self, path: str | Path) -> None:
        """
        Удаляем файл.
        """
        path = self.work_dir / path
        await aos.remove(path)



"""
Модуль, содержащий репозиторий для выполнения CRUD операций над моделями в базе данных.
"""

from __future__ import annotations

import contextlib
import uuid
from collections.abc import Callable, Iterable, Sequence
from itertools import groupby
from typing import Any, Generic, Self, cast, get_args

import sqlalchemy as sa
from sqlalchemy.dialects.postgresql import insert
from sqlalchemy.exc import IntegrityError
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.orm import selectin_polymorphic
from sqlalchemy.sql.expression import func

from fast_clean.db import SessionManagerProtocol
from fast_clean.enums import ModelActionEnum
from fast_clean.exceptions import (
    ModelIntegrityError,
    ModelNotFoundError,
    SortingFieldNotFoundError,
)
from fast_clean.schemas import PaginationResultSchema, PaginationSchema

from .type_vars import (
    CreateSchemaBaseType,
    CreateSchemaIntType,
    CreateSchemaType,
    IdType,
    ModelBaseType,
    ModelIntType,
    ModelType,
    ReadSchemaBaseType,
    ReadSchemaIntType,
    ReadSchemaType,
    UpdateSchemaBaseType,
    UpdateSchemaIntType,
    UpdateSchemaType,
)


class DbCrudRepositoryBase(
    Generic[
        ModelBaseType,
        ReadSchemaBaseType,
        CreateSchemaBaseType,
        UpdateSchemaBaseType,
        IdType,
    ]
):
    """
    Базовый репозиторий для выполнения CRUD операций над моделями в базе данных.
    """

    __abstract__: bool = True

    __orig_bases__: tuple[
        type[
            DbCrudRepositoryBase[ModelBaseType, ReadSchemaBaseType, CreateSchemaBaseType, UpdateSchemaBaseType, IdType]
        ]
    ]
    __subtypes__: Sequence[
        tuple[type[ModelBaseType], type[ReadSchemaBaseType], type[CreateSchemaBaseType], type[UpdateSchemaBaseType]]
    ]

    model_types: set[type[ModelBaseType]] = set()
    model_subtypes: set[type[ModelBaseType]] = set()
    model_types_mapping: dict[type[ModelBaseType], type[ReadSchemaBaseType]]
    create_models_mapping: dict[type[CreateSchemaBaseType], type[ModelBaseType]]
    update_models_mapping: dict[type[UpdateSchemaBaseType], type[ModelBaseType]]
    model_identities_mapping: dict[Any, type[ModelBaseType]] = {}

    model_type: type[ModelBaseType]

    def __init__(self, session_manager: SessionManagerProtocol):
        if self.__dict__.get('__abstract__', False):
            raise TypeError(f"Can't instantiate abstract class {type(self).__name__}")
        self.session_manager = session_manager

    def __init_subclass__(cls) -> None:
        """
        Инициализируем класс.

        Получаем используемую модель SQLAlchemy и схему Pydantic из базового типа.
        """
        if cls.__dict__.get('__abstract__', False):
            return super().__init_subclass__()

        base_repository_generic = next(
            (
                base
                for base in getattr(cls, '__orig_bases__', [])
                if issubclass(getattr(base, '__origin__', base), DbCrudRepositoryBase)
            ),
            None,
        )
        if not base_repository_generic:
            raise ValueError('Repository must be implemented by DbCrudRepositoryBase')

        if not hasattr(cls, '__subtypes__'):
            cls.__subtypes__ = []

        cls.model_subtypes = {st[0] for st in cls.__subtypes__}

        cls.model_types_mapping = {}
        cls.create_models_mapping = {}
        cls.update_models_mapping = {}
        cls.model_identities_mapping = {}
        types: Sequence[
            tuple[type[ModelBaseType], type[ReadSchemaBaseType], type[CreateSchemaBaseType], type[UpdateSchemaBaseType]]
        ] = [*cls.__subtypes__, get_args(base_repository_generic)[:4]]
        for model_type, read_schema_type, create_schema_type, update_schema_type in types:
            cls.model_types.add(model_type)
            cls.model_types_mapping[model_type] = read_schema_type
            cls.create_models_mapping[create_schema_type] = model_type
            cls.update_models_mapping[update_schema_type] = model_type
            cls.model_identities_mapping[model_type.__mapper__.polymorphic_identity] = model_type

        cls.model_type, *_ = cast(
            tuple[
                type[ModelBaseType],
                type[ReadSchemaBaseType],
                type[CreateSchemaBaseType],
                type[UpdateSchemaBaseType],
            ],
            get_args(base_repository_generic),
        )

        return super().__init_subclass__()

    async def get(self: Self, id: IdType) -> ReadSchemaBaseType:
        """
        Получаем модель по идентификатору.
        """
        async with self.session_manager.get_session() as s:
            statement = self.select().where(self.model_type.id == id)
            model = (await s.execute(statement)).scalar_one_or_none()
            if model is None:
                raise ModelNotFoundError(self.model_type, model_id=id)
            return self.model_validate(model)

    async def get_or_none(self: Self, id: IdType) -> ReadSchemaBaseType | None:
        """
        Получаем модель или None по идентификатору.
        """
        with contextlib.suppress(ModelNotFoundError):
            return await self.get(id)
        return None

    async def get_by_ids(self: Self, ids: Sequence[IdType], *, exact: bool = False) -> list[ReadSchemaBaseType]:
        """
        Получаем список моделей по идентификаторам.
        """
        async with self.session_manager.get_session() as s:
            statement = self.select().where(self.model_type.id.in_(ids))
            models = (await s.execute(statement)).scalars().all()
            self.check_get_by_ids_exact(ids, models, exact)
            return [self.model_validate(model) for model in models]

    async def get_all(self: Self) -> list[ReadSchemaBaseType]:
        """
        Получаем все модели.
        """
        async with self.session_manager.get_session() as s:
            statement = self.select()
            models = (await s.execute(statement)).scalars().all()
            return [self.model_validate(model) for model in models]

    async def paginate(
        self: Self,
        pagination: PaginationSchema,
        user: Any,
        policies: list[str],
        *,
        search: str | None = None,
        search_by: Iterable[str] | None = None,
        sorting: Iterable[str] | None = None,
    ) -> PaginationResultSchema[ReadSchemaBaseType]:
        """
        Получаем список моделей с пагинацией, поиском и сортировкой.
        """
        return await self.paginate_with_filter(
            pagination,
            user,
            policies,
            search=search,
            search_by=search_by,
            sorting=sorting,
        )

    async def create(self: Self, create_object: CreateSchemaBaseType) -> ReadSchemaBaseType:
        """
        Создаем модель.
        """
        async with self.session_manager.get_session() as s:
            try:
                model_type = self.create_models_mapping[type(create_object)]
                create_dict = self.dump_create_object(create_object)
                return (await self.bulk_create_with_model_type(model_type, [create_dict], s))[0]
            except IntegrityError as integrity_error:
                raise ModelIntegrityError(self.model_type, ModelActionEnum.INSERT) from integrity_error

    async def bulk_create(self: Self, create_objects: list[CreateSchemaBaseType]) -> list[ReadSchemaBaseType]:
        """
        Создаем несколько моделей.
        """
        if len(create_objects) == 0:
            return []
        async with self.session_manager.get_session() as s:
            try:
                created_models: list[ReadSchemaBaseType] = []
                for model_type, type_create_objects in groupby(
                    create_objects, key=lambda co: self.create_models_mapping[type(co)]
                ):
                    create_dicts = [self.dump_create_object(create_object) for create_object in type_create_objects]
                    created_models.extend(await self.bulk_create_with_model_type(model_type, create_dicts, s))
                return created_models
            except IntegrityError as integrity_error:
                raise ModelIntegrityError(self.model_type, ModelActionEnum.INSERT) from integrity_error

    async def update(self: Self, update_object: UpdateSchemaBaseType) -> ReadSchemaBaseType:
        """
        Обновляем модель.
        """
        async with self.session_manager.get_session() as s:
            try:
                model_type = self.update_models_mapping[type(update_object)]
                update_dict = update_object.model_dump(exclude_unset=True)
                return await self.update_with_model_type(model_type, update_dict, s)
            except IntegrityError as integrity_error:
                raise ModelIntegrityError(self.model_type, ModelActionEnum.UPDATE) from integrity_error

    async def bulk_update(self: Self, update_objects: list[UpdateSchemaBaseType]) -> None:
        """
        Обновляем несколько моделей.
        """
        if len(update_objects) == 0:
            return
        async with self.session_manager.get_session() as s:
            try:
                for model_type, type_update_objects in groupby(
                    update_objects, key=lambda co: self.update_models_mapping[type(co)]
                ):
                    update_dicts = [update_object.model_dump() for update_object in type_update_objects]
                    await self.bulk_update_with_model_type(model_type, update_dicts, s)
            except IntegrityError as integrity_error:
                raise ModelIntegrityError(self.model_type, ModelActionEnum.UPDATE) from integrity_error

    async def upsert(self: Self, create_object: CreateSchemaBaseType) -> ReadSchemaBaseType:
        """
        Создаем или обновляем модель.
        """
        async with self.session_manager.get_session() as s:
            try:
                model_type = self.create_models_mapping[type(create_object)]
                create_dict = self.dump_create_object(create_object)
                return await self.upsert_with_model_type(model_type, create_dict, s)
            except IntegrityError as integrity_error:
                raise ModelIntegrityError(self.model_type, ModelActionEnum.UPSERT) from integrity_error

    async def delete(self: Self, ids: Sequence[IdType]) -> None:
        """
        Удаляем модели.
        """
        if len(ids) == 0:
            return
        async with self.session_manager.get_session() as s:
            try:
                model_types: list[type[ModelBaseType]] = [self.model_type]
                if self.model_type.__mapper__.polymorphic_on is not None:
                    types_statement = (
                        sa.select(self.model_type.__mapper__.polymorphic_on)
                        .where(self.model_type.id.in_(ids))
                        .distinct()
                    )
                    types = (await s.execute(types_statement)).scalars().all()
                    for t in types:
                        model_types.append(self.model_identities_mapping[t])
                for model_type in model_types[::-1]:
                    statement = sa.delete(model_type).where(model_type.id.in_(ids))
                    await s.execute(statement)
            except IntegrityError as integrity_error:
                raise ModelIntegrityError(self.model_type, ModelActionEnum.DELETE) from integrity_error

    @classmethod
    def select(cls) -> sa.Select[tuple[ModelBaseType]]:
        """
        Выбираем базовую модели или наследника со всеми полями при наличии.
        """
        statement = sa.select(cls.model_type)
        if cls.model_subtypes:
            return statement.options(selectin_polymorphic(cls.model_type, cls.model_subtypes))
        return statement

    @classmethod
    def model_validate(cls, model: ModelBaseType) -> ReadSchemaBaseType:
        """
        Приводим модель к схеме.
        """
        read_schema_type = cls.model_types_mapping[type(model)]
        return cast(
            ReadSchemaBaseType,
            read_schema_type.model_validate(model, from_attributes=True),
        )

    @classmethod
    async def bulk_create_with_model_type(
        cls, model_type: type[ModelBaseType], create_dicts: list[dict[str, Any]], session: AsyncSession
    ) -> list[ReadSchemaBaseType]:
        """
        Создаем модели с помощью типа.
        """
        parent_dicts = await cls.bulk_create_parent_model(model_type, create_dicts, session)
        values: list[dict[str, Any]] = []
        for create_dict, parent_dict in zip(create_dicts, parent_dicts, strict=True):
            value = {k: v for k, v in create_dict.items() if k in model_type.__table__.columns}
            if 'id' in parent_dict:
                value['id'] = parent_dict['id']
            values.append(value)
        statement = sa.insert(model_type).values(values).returning(*model_type.__table__.columns.values())
        model_dicts = (await session.execute(statement)).mappings().all()
        read_schema_type = cls.model_types_mapping[model_type]
        return [
            cast(
                ReadSchemaBaseType,
                read_schema_type.model_validate({**parent_dict, **model_dict}),
            )
            for parent_dict, model_dict in zip(parent_dicts, model_dicts, strict=True)
        ]

    @classmethod
    async def bulk_create_parent_model(
        cls, model_type: type[ModelBaseType], create_dicts: list[dict[str, Any]], session: AsyncSession
    ) -> list[dict[str, Any]]:
        """
        Создаем родительские модели с помощью типа.
        """
        parent_model_type = cls.get_parent_model_type(model_type)
        if parent_model_type is None:
            return [{} for _ in create_dicts]
        return [
            ps.model_dump() for ps in await cls.bulk_create_with_model_type(parent_model_type, create_dicts, session)
        ]

    @classmethod
    async def update_with_model_type(
        cls, model_type: type[ModelBaseType], update_dict: dict[str, Any], session: AsyncSession
    ) -> ReadSchemaBaseType:
        """
        Обновляем модель с помощью типа.
        """
        parent_dict = await cls.update_parent_model(model_type, update_dict, session)
        statement = (
            sa.update(model_type)
            .where(model_type.id == update_dict['id'])
            .values({k: v for k, v in update_dict.items() if k in model_type.__table__.columns if k != 'id'})
            .returning(*model_type.__table__.columns.values())
        )
        model_dict = (await session.execute(statement)).mappings().one()
        read_schema_type = cls.model_types_mapping[model_type]
        return cast(
            ReadSchemaBaseType,
            read_schema_type.model_validate({**parent_dict, **model_dict}),
        )

    @classmethod
    async def update_parent_model(
        cls, model_type: type[ModelBaseType], update_dict: dict[str, Any], session: AsyncSession
    ) -> dict[str, Any]:
        """
        Обновляем родительскую модель с помощью типа.
        """
        parent_model_type = cls.get_parent_model_type(model_type)
        if parent_model_type is None:
            return {}
        return (await cls.update_with_model_type(parent_model_type, update_dict, session)).model_dump()

    @classmethod
    async def bulk_update_with_model_type(
        cls, model_type: type[ModelBaseType], update_dicts: list[dict[str, Any]], session: AsyncSession
    ) -> None:
        """
        Создаем модели с помощью типа.
        """
        await cls.bulk_update_parent_model(model_type, update_dicts, session)
        values: list[dict[str, Any]] = []
        for update_dict in update_dicts:
            values.append({k: v for k, v in update_dict.items() if k in model_type.__table__.columns})
        statement = sa.update(model_type)
        await session.execute(statement, values)

    @classmethod
    async def bulk_update_parent_model(
        cls, model_type: type[ModelBaseType], update_dicts: list[dict[str, Any]], session: AsyncSession
    ) -> None:
        """
        Обновляем родительские модели с помощью типа.
        """
        parent_model_type = cls.get_parent_model_type(model_type)
        if parent_model_type is not None:
            await cls.bulk_update_with_model_type(parent_model_type, update_dicts, session)

    @classmethod
    async def upsert_with_model_type(
        cls, model_type: type[ModelBaseType], create_dict: dict[str, Any], session: AsyncSession
    ) -> ReadSchemaBaseType:
        """
        Создаем или обновляем модель с помощью типа.
        """
        parent_dict = await cls.upsert_parent_model(model_type, create_dict, session)
        primary_keys = {key.name for key in cast(Any, sa.inspect(model_type)).primary_key}
        values = {k: v for k, v in create_dict.items() if k in model_type.__table__.columns}
        statement = (
            insert(model_type)
            .values(values)
            .on_conflict_do_update(
                index_elements=primary_keys,
                set_={k: v for k, v in values.items() if k not in primary_keys},
            )
            .returning(*model_type.__table__.columns.values())
        )
        model_dict = (await session.execute(statement)).mappings().one()
        read_schema_type = cls.model_types_mapping[model_type]
        return cast(
            ReadSchemaBaseType,
            read_schema_type.model_validate({**parent_dict, **model_dict}),
        )

    @classmethod
    async def upsert_parent_model(
        cls, model_type: type[ModelBaseType], create_dict: dict[str, Any], session: AsyncSession
    ) -> dict[str, Any]:
        """
        Создаем или обновляем родительскую модель с помощью типа.
        """
        parent_model_type = cls.get_parent_model_type(model_type)
        if parent_model_type is None:
            return {}
        return (await cls.upsert_with_model_type(parent_model_type, create_dict, session)).model_dump()

    @staticmethod
    def dump_create_object(create_object: CreateSchemaBaseType) -> dict[str, Any]:
        """
        Создаем словарь для схемы создания модели.
        """
        create_dict = create_object.model_dump()
        if create_dict['id'] is None:
            del create_dict['id']
        return create_dict

    @classmethod
    def get_parent_model_type(cls, model_type: type[ModelBaseType]) -> type[ModelBaseType] | None:
        """
        Получаем тип родительской модели.
        """
        if not model_type.__bases__ or model_type.__bases__[0] not in cls.model_types:
            return None
        return model_type.__bases__[0]

    @classmethod
    def check_get_by_ids_exact(cls, ids: Sequence[IdType], models: Sequence[ModelBaseType], exact: bool) -> None:
        """
        Проверяем, что по идентификаторам получены все модели.
        """
        if exact and len(ids) != len(models):
            raise ModelNotFoundError(
                cls.model_type,
                model_id=set(ids) - {cast(IdType, model.id) for model in models},
            )

    async def paginate_with_filter(
        self: Self,
        pagination: PaginationSchema,
        user: Any,
        policies: list[str],
        *,
        search: str | None = None,
        search_by: Iterable[str] | None = None,
        sorting: Iterable[str] | None = None,
        select_filter: Callable[[sa.Select[tuple[ModelBaseType]]], sa.Select[tuple[ModelBaseType]]] | None = None,
    ) -> PaginationResultSchema[ReadSchemaBaseType]:
        """
        Получаем список моделей с пагинацией, поиском, сортировкой и фильтрами.
        """
        if len(policies) == 0:
            return PaginationResultSchema(objects=[], count=0)
        search_by = search_by or []
        sorting = sorting or []
        async with self.session_manager.get_session() as s:
            statement = self.select()
            if select_filter:
                statement = select_filter(statement)
            if search:
                search_where: sa.ColumnElement[Any] = sa.false()
                for sb in search_by:
                    search_where = sa.or_(search_where, getattr(self.model_type, sb).ilike(f'%{search}%'))
                statement = statement.where(search_where)
            order_by_expr = self.get_order_by_expr(sorting)
            models = (
                (await s.execute(statement.limit(pagination.limit).offset(pagination.offset).order_by(*order_by_expr)))
                .scalars()
                .all()
            )
            objects = [self.model_validate(model) for model in models]
            count_statement = statement.with_only_columns(func.count(self.model_type.id))
            count = (await s.execute(count_statement)).scalar_one()
            return PaginationResultSchema(count=count, objects=objects)

    def get_order_by_expr(self: Self, sorting: Iterable[str]) -> list[sa.UnaryExpression]:
        """
        Получаем выражение сортировки.
        """
        order_by_expr: list[sa.UnaryExpression] = []
        for st in sorting:
            try:
                if st[0] == '-':
                    order_by_expr.append(getattr(self.model_type, st[1:]).desc())
                else:
                    order_by_expr.append(getattr(self.model_type, st))
            except AttributeError as attribute_error:
                raise SortingFieldNotFoundError(st) from attribute_error
        return order_by_expr


class DbCrudRepositoryInt(
    DbCrudRepositoryBase[
        ModelIntType,
        ReadSchemaIntType,
        CreateSchemaIntType,
        UpdateSchemaIntType,
        int,
    ],
    Generic[
        ModelIntType,
        ReadSchemaIntType,
        CreateSchemaIntType,
        UpdateSchemaIntType,
    ],
):
    """
    Репозиторий для выполнения CRUD операций над моделями старого типа в базе данных.
    """

    __abstract__ = True


class DbCrudRepository(
    DbCrudRepositoryBase[ModelType, ReadSchemaType, CreateSchemaType, UpdateSchemaType, uuid.UUID],
    Generic[ModelType, ReadSchemaType, CreateSchemaType, UpdateSchemaType],
):
    """
    Репозиторий для выполнения CRUD операций над моделями нового типа в базе данных.
    """

    __abstract__ = True



"""
Модуль, содержащий репозиторий кеша с помощью Redis.
"""

from typing import Self

from fastapi_cache.backends.redis import RedisBackend
from overrides import override

from redis.asyncio.client import Redis


class RedisCacheRepository(RedisBackend):
    """
    Репозиторий кеша с помощью Redis.
    """

    def __init__(self, redis: Redis):
        super().__init__(redis)
        self.redis: Redis

    @override(check_signature=False)
    async def set(self: Self, key: str, value: str, expire: int | None = None, nx: bool = False) -> None:
        """
        Устанавливаем значение.
        """
        await self.redis.set(key, value, ex=expire, nx=nx)

    async def incr(self: Self, key: str, amount: int = 1) -> int:
        """
        Инкремент значения.
        """
        return await self.redis.incr(key, amount)

    async def decr(self: Self, key: str, amount: int = 1) -> int:
        """
        Декремент значения.
        """
        return await self.redis.decr(key, amount)

    async def clear(self, namespace: str | None = None, key: str | None = None) -> int:
        """
        Удаляем значение.

        Родительский метод работает не правильно и не подсчитывает количество удаленных записей.
        https://github.com/long2ice/fastapi-cache/issues/241
        """
        if namespace:
            cursor = 0
            removed = 0
            while True:
                cursor, keys = await self.redis.scan(cursor, match=f'{namespace}:*', count=500)
                removed += await self.redis.delete(*keys)
                if cursor == 0:
                    return removed
        elif key:
            return await self.redis.delete(key)
        return 0



"""
Модуль, содержащий классы для потокового чтения данных.
"""

from collections.abc import AsyncIterator
from typing import Protocol, Self

from aiofiles.threadpool.binary import AsyncBufferedReader
from aiohttp import ClientResponse

READ_SIZE = 5 * 1024 * 1024


class StreamReadSyncProtocol(Protocol):
    """
    Синхронный протокол потокового чтения данных.
    """

    def read(self: Self, size: int | None = READ_SIZE) -> bytes:
        """
        Читаем данные.
        """
        ...


class StreamReadAsyncProtocol(Protocol):
    """
    Асинхронный протокол потокового чтения данных.
    """

    async def read(self: Self, size: int | None = READ_SIZE) -> bytes:
        """
        Читаем данные.
        """
        ...


StreamReadProtocol = StreamReadAsyncProtocol | StreamReadSyncProtocol


class StreamReaderProtocol(Protocol):
    """
    Протокол потокового чтения данных с контекстным менеджером.
    """

    async def read(self: Self, size: int = READ_SIZE) -> bytes:
        """
        Читаем данные.
        """
        ...

    def __aiter__(self: Self) -> AsyncIterator[bytes]:
        """
        Заходим в контекстный менеджер.
        """
        ...

    async def __anext__(self: Self) -> bytes:
        """
        Читаем следующую порцию данных.
        """
        ...


class AiofilesStreamReader:
    """
    Реализация потокового чтения данных для библиотеки `aiofiles`.
    """

    def __init__(self, reader: AsyncBufferedReader) -> None:
        self.reader = reader

    async def read(self: Self, size: int = READ_SIZE) -> bytes:
        """
        Читаем данные.
        """
        return await self.reader.read(size)

    def __aiter__(self: Self) -> AsyncIterator[bytes]:
        """
        Заходим в контекстный менеджер.
        """
        return self

    async def __anext__(self: Self) -> bytes:
        """
        Читаем следующую порцию данных.
        """
        chunk = await self.reader.read(READ_SIZE)
        if chunk:
            return chunk
        raise StopAsyncIteration()


class AiohttpStreamReader:
    """
    Реализация потокового чтения данных для библиотеки `aiohttp`.
    """

    def __init__(self, response: ClientResponse) -> None:
        self.response = response

    async def read(self: Self, size: int = READ_SIZE) -> bytes:
        """
        Читаем данные.
        """
        return await self.response.content.read(size)

    def __aiter__(self: Self) -> AsyncIterator[bytes]:
        """
        Заходим в контекстный менеджер.
        """
        return self

    async def __anext__(self: Self) -> bytes:
        """
        Читаем следующую порцию данных.
        """
        chunk = await self.response.content.read(READ_SIZE)
        if chunk:
            return chunk
        raise StopAsyncIteration()



"""
Модуль, содержащий репозиторий для выполнения CRUD операций над моделями в памяти.
"""

import contextlib
import uuid
from abc import ABC, abstractmethod
from collections.abc import Iterable, Sequence
from itertools import groupby
from typing import Any, Callable, Generic, Self, cast, get_args

from fast_clean.enums import ModelActionEnum
from fast_clean.exceptions import ModelIntegrityError, ModelNotFoundError
from fast_clean.schemas import PaginationResultSchema, PaginationSchema

from .type_vars import (
    CreateSchemaBaseType,
    CreateSchemaIntType,
    CreateSchemaType,
    IdType,
    ReadSchemaBaseType,
    ReadSchemaIntType,
    ReadSchemaType,
    UpdateSchemaBaseType,
    UpdateSchemaIntType,
    UpdateSchemaType,
)


class InMemoryCrudRepositoryBase(ABC, Generic[ReadSchemaBaseType, CreateSchemaBaseType, UpdateSchemaBaseType, IdType]):
    """
    Базовый репозиторий для выполнения CRUD операций над моделями в памяти.
    """

    __abstract__: bool = True

    __orig_bases__: 'tuple[type[InMemoryCrudRepositoryBase[ReadSchemaBaseType, CreateSchemaBaseType, UpdateSchemaBaseType, IdType]]]'
    __subtypes__: Sequence[tuple[type[ReadSchemaBaseType], type[CreateSchemaBaseType], type[UpdateSchemaBaseType]]]

    create_to_read_schemas_mapping: dict[type[CreateSchemaBaseType], type[ReadSchemaBaseType]]
    create_to_update_schemas_mapping: dict[type[CreateSchemaBaseType], type[UpdateSchemaBaseType]]
    update_to_read_schemas_mapping: dict[type[UpdateSchemaBaseType], type[ReadSchemaBaseType]]

    read_schema_type: type[ReadSchemaBaseType]

    def __init__(self) -> None:
        if self.__dict__.get('__abstract__', False):
            raise TypeError(f"Can't instantiate abstract class {type(self).__name__}")
        self.models: dict[IdType, ReadSchemaBaseType] = {}

    def __init_subclass__(cls) -> None:
        """
        Инициализируем класс.

        Получаем схемы Pydantic из базового класса.
        """
        if cls.__dict__.get('__abstract__', False):
            return super().__init_subclass__()

        base_repository_generic = next(
            (
                base
                for base in getattr(cls, '__orig_bases__', [])
                if issubclass(getattr(base, '__origin__', base), InMemoryCrudRepositoryBase)
            ),
            None,
        )
        if not base_repository_generic:
            raise ValueError('Repository must be implemented by InMemoryCrudRepositoryBase')

        if not hasattr(cls, '__subtypes__'):
            cls.__subtypes__ = []

        cls.create_to_read_schemas_mapping = {}
        cls.create_to_update_schemas_mapping = {}
        cls.update_to_read_schemas_mapping = {}
        types: Sequence[tuple[type[ReadSchemaBaseType], type[CreateSchemaBaseType], type[UpdateSchemaBaseType]]] = [
            *cls.__subtypes__,
            get_args(base_repository_generic)[:3],
        ]
        for read_schema_type, create_schema_type, update_schema_type in types:
            cls.create_to_read_schemas_mapping[create_schema_type] = read_schema_type
            cls.create_to_update_schemas_mapping[create_schema_type] = update_schema_type
            cls.update_to_read_schemas_mapping[update_schema_type] = read_schema_type

        cls.read_schema_type, *_ = cast(
            tuple[
                type[ReadSchemaBaseType],
                type[CreateSchemaBaseType],
                type[UpdateSchemaBaseType],
            ],
            get_args(base_repository_generic),
        )
        return super().__init_subclass__()

    @abstractmethod
    def generate_id(self: Self) -> IdType:
        """
        Генерируем идентификатор.
        """
        ...

    def get_model_name(self: Self, read_schema_type: type[ReadSchemaBaseType] | None = None) -> str:
        """
        Получаем название модели.
        """
        read_schema_type = read_schema_type or self.read_schema_type
        return read_schema_type.__name__.replace('Schema', '')

    async def get(self: Self, id: IdType) -> ReadSchemaBaseType:
        """
        Получаем модель по идентификатору.
        """
        model = self.models.get(id)
        if model is None:
            raise ModelNotFoundError(self.get_model_name(), model_id=id)
        return model

    async def get_or_none(self: Self, id: IdType) -> ReadSchemaBaseType | None:
        """
        Получаем модель или None по идентификатору.
        """
        with contextlib.suppress(ModelNotFoundError):
            return await self.get(id)
        return None

    async def get_by_ids(self: Self, ids: Sequence[IdType], *, exact: bool = False) -> list[ReadSchemaBaseType]:
        """
        Получаем список моделей по идентификаторам.
        """
        models: list[ReadSchemaBaseType] = []
        for id in ids:
            model = self.models.get(id)
            if model is not None:
                models.append(model)
        self.check_get_by_ids_exact(ids, models, exact)
        return models

    async def get_all(self: Self) -> list[ReadSchemaBaseType]:
        """
        Получаем все модели.
        """
        return list(self.models.values())

    async def paginate(
        self: Self,
        pagination: PaginationSchema,
        user: Any,
        policies: list[str],
        *,
        search: str | None = None,
        search_by: Iterable[str] | None = None,
        sorting: Iterable[str] | None = None,
    ) -> PaginationResultSchema[ReadSchemaBaseType]:
        """
        Получаем список моделей с пагинацией, поиском и сортировкой.
        """
        return self.paginate_with_filter(
            pagination,
            user,
            policies,
            search=search,
            search_by=search_by,
            sorting=sorting,
        )

    async def create(self: Self, create_object: CreateSchemaBaseType) -> ReadSchemaBaseType:
        """
        Создаем модель.
        """
        model = self.make_model(create_object)
        self.models[cast(IdType, model.id)] = model
        return model

    async def bulk_create(self: Self, create_objects: list[CreateSchemaBaseType]) -> list[ReadSchemaBaseType]:
        """
        Создаем несколько моделей.
        """
        models: list[ReadSchemaBaseType] = []
        for create_object in create_objects:
            models.append(self.make_model(create_object))
        for model in models:
            self.models[cast(IdType, model.id)] = model
        return models

    async def update(self: Self, update_object: UpdateSchemaBaseType) -> ReadSchemaBaseType:
        """
        Обновляем модель.
        """
        read_schema_type = self.update_to_read_schemas_mapping[type(update_object)]
        model = self.models.get(cast(IdType, update_object.id))
        if model is None:
            raise ModelNotFoundError(self.get_model_name(read_schema_type), model_id=update_object.id)
        model = cast(
            ReadSchemaBaseType,
            read_schema_type.model_validate(
                {
                    **model.model_dump(),
                    **update_object.model_dump(exclude={'id'}, exclude_unset=True),
                }
            ),
        )
        self.models[cast(IdType, model.id)] = model
        return model

    async def bulk_update(self: Self, update_objects: list[UpdateSchemaBaseType]) -> None:
        """
        Обновляем несколько моделей.
        """
        for update_object in update_objects:
            await self.update(update_object)

    async def upsert(self: Self, create_object: CreateSchemaBaseType) -> ReadSchemaBaseType:
        """
        Создаем или обновляем модель.
        """
        if create_object.id is None or create_object.id not in self.models:
            return await self.create(create_object)
        update_schema_type = self.create_to_update_schemas_mapping[type(create_object)]
        await self.update(
            cast(
                UpdateSchemaBaseType,
                update_schema_type.model_validate(create_object.model_dump()),
            )
        )
        return self.models[cast(IdType, create_object.id)]

    async def delete(self: Self, ids: Sequence[IdType]) -> None:
        """
        Удаляем модели.
        """
        for id in ids:
            if id in self.models:
                del self.models[id]

    def check_get_by_ids_exact(
        self: Self,
        ids: Sequence[IdType],
        models: Sequence[ReadSchemaBaseType],
        exact: bool,
    ) -> None:
        """
        Проверяем, что по идентификаторам получены все модели.
        """
        if exact and len(ids) != len(models):
            raise ModelNotFoundError(
                self.get_model_name(),
                model_id=set(ids) - {cast(IdType, model.id) for model in models},
            )

    def paginate_with_filter(
        self: Self,
        pagination: PaginationSchema,
        user: Any,
        policies: list[str],
        *,
        search: str | None = None,
        search_by: Iterable[str] | None = None,
        sorting: Iterable[str] | None = None,
        select_filter: Callable[[ReadSchemaBaseType], bool] | None = None,
    ) -> PaginationResultSchema[ReadSchemaBaseType]:
        """
        Получаем список моделей с пагинацией, поиском, сортировкой и фильтрами.
        """
        if len(policies) == 0:
            return PaginationResultSchema(objects=[], count=0)
        search_by = search_by or []
        sorting = sorting or []
        models = list(filter(select_filter, self.models.values()) if select_filter else self.models.values())
        if search:
            search_models: list[ReadSchemaBaseType] = []
            for model in models:
                for sb in search_by:
                    if search in getattr(model, sb):
                        search_models.append(model)
            models = search_models
        models = self.sort(models, sorting)
        return PaginationResultSchema(
            objects=models[pagination.offset : pagination.offset + pagination.limit],
            count=len(models),
        )

    @classmethod
    def sort(cls, models: list[ReadSchemaBaseType], sorting: Iterable[str]) -> list[ReadSchemaBaseType]:
        """
        Сортируем модели.
        """
        if not sorting:
            return models
        st, *sorting = sorting
        if st[0] == '-':
            sorted_models = sorted(models, key=lambda model: getattr(model, st[1:]), reverse=True)
        else:
            sorted_models = sorted(models, key=lambda model: getattr(model, st))
        result: list[ReadSchemaBaseType] = []
        for _, group_models in groupby(sorted_models, lambda model: getattr(model, st[1:] if st[0] == '-' else st)):
            result.extend(cls.sort(list(group_models), sorting))
        return result

    def make_model(self: Self, create_object: CreateSchemaBaseType) -> ReadSchemaBaseType:
        """
        Создаем модель без сохранения.
        """
        create_dict = create_object.model_dump()
        if create_dict['id'] is None:
            create_dict['id'] = self.generate_id()
        read_schema_type = self.create_to_read_schemas_mapping[type(create_object)]
        model = cast(ReadSchemaBaseType, read_schema_type.model_validate(create_dict))
        if model.id in self.models:
            raise ModelIntegrityError(self.get_model_name(read_schema_type), ModelActionEnum.INSERT)
        return model


class InMemoryCrudRepositoryInt(
    InMemoryCrudRepositoryBase[
        ReadSchemaIntType,
        CreateSchemaIntType,
        UpdateSchemaIntType,
        int,
    ],
    Generic[
        ReadSchemaIntType,
        CreateSchemaIntType,
        UpdateSchemaIntType,
    ],
):
    """
    Репозиторий для выполнения CRUD операций над моделями старого типа в памяти.
    """

    __abstract__ = True

    def __init__(self) -> None:
        super().__init__()
        self.ids_counter = 0

    def generate_id(self: Self) -> int:
        """
        Генерируем идентификатор.
        """
        current_id = self.ids_counter
        self.ids_counter += 1
        return current_id


class InMemoryCrudRepository(
    InMemoryCrudRepositoryBase[
        ReadSchemaType,
        CreateSchemaType,
        UpdateSchemaType,
        uuid.UUID,
    ],
    Generic[
        ReadSchemaType,
        CreateSchemaType,
        UpdateSchemaType,
    ],
):
    """
    Репозиторий для выполнения CRUD операций над моделями нового типа в памяти.
    """

    __abstract__ = True

    def generate_id(self: Self) -> uuid.UUID:
        """
        Генерируем идентификатор.
        """
        return uuid.uuid4()



"""
Модуль, содержащий тесты исключений.
"""

from .exceptions import CustomTestError


class TestBusinessLogicException:
    """
    Тесты базового исключения бизнес-логики.
    """

    EXPECTED_TYPE = 'custom_test'
    EXPECTED_MSG = 'Тестовое сообщение'
    EXPECTED_PARENT_MSG = 'Родительское сообщение'

    @classmethod
    def test_get_schema_not_debug(cls) -> None:
        """
        Тестируем метод `get_schema` без отладки.
        """
        try:
            raise CustomTestError()
        except CustomTestError as test_error:
            schema = test_error.get_schema(False)
            assert schema.type == cls.EXPECTED_TYPE
            assert schema.msg == cls.EXPECTED_MSG
            assert schema.traceback is None

    @classmethod
    def test_get_schema_debug(cls) -> None:
        """
        Тестируем метод `get_schema` с отладкой.
        """
        try:
            raise CustomTestError() from Exception(cls.EXPECTED_PARENT_MSG)
        except CustomTestError as test_error:
            schema = test_error.get_schema(True)
            assert schema.type == cls.EXPECTED_TYPE
            assert schema.msg == cls.EXPECTED_MSG
            assert schema.traceback is not None
            assert CustomTestError.__name__ in schema.traceback
            assert cls.EXPECTED_MSG in schema.traceback
            assert Exception.__name__ in schema.traceback
            assert cls.EXPECTED_PARENT_MSG in schema.traceback



"""
Модуль, содержащий сервис транзакций.
"""

from contextlib import asynccontextmanager
from typing import AsyncIterator

import sqlalchemy as sa
from sqlalchemy.ext.asyncio import AsyncSession


class TransactionService:
    """
    Реализация сервиса транзакций.
    """

    def __init__(self, session: AsyncSession) -> None:
        self.session = session

    @asynccontextmanager
    async def begin(self, immediate: bool = True) -> AsyncIterator[None]:
        """
        Начинаем транзакцию.
        """
        async with self.session.begin():
            if immediate:
                await self.session.execute(sa.text('SET CONSTRAINTS ALL IMMEDIATE'))
            yield



"""
Модуль, содержащий репозиторий кеша в памяти.
"""

from typing import Self

from fastapi_cache.backends.inmemory import InMemoryBackend, Value
from overrides import override


class InMemoryCacheRepository(InMemoryBackend):
    """
    Репозиторий кеша в памяти.
    """

    @override(check_signature=False)
    def _get(self, key: str) -> Value | None:
        """
        Получаем внутреннее значение.

        Родительский метод работает неправильно, т.к. не рассчитан на правильную логику метода `set`.
        """
        v = self._store.get(key)
        if v:
            if v.ttl_ts == -1 or v.ttl_ts >= self._now:
                return v
            else:
                del self._store[key]
        return None

    @override(check_signature=False)
    async def set(self: Self, key: str, value: str, expire: int | None = None, nx: bool = False) -> None:
        """
        Устанавливаем значение.

        Родительский метод работает неправильно, т.к. при отсутствии `expire` должно устанавливаться `-1`.
        Старая логика приводит к тому, что метод работает не так, как `RedisBackend`.
        """
        async with self._lock:
            ttl_ts = self._now + expire if expire is not None else -1
            existing_value = self._get(key)
            if not nx or existing_value is None:
                self._store[key] = Value(value, ttl_ts)  # type: ignore

    async def incr(self: Self, key: str, amount: int = 1) -> int:
        """
        Инкремент значения.
        """
        v = self._get(key)
        if v is None:
            n_value = amount
            await self.set(key, str(amount))
            return n_value
        n_value = int(v.data) + amount
        await self.set(key, str(n_value))
        return n_value

    async def decr(self: Self, key: str, amount: int = 1) -> int:
        """
        Декремент значения.
        """
        return await self.incr(key, -amount)



"""
Модуль, содержащий функционал, связанный с брокером сообщений.
"""

from collections.abc import AsyncIterator
from contextlib import asynccontextmanager
from functools import partial
from typing import Callable, Self

from faststream import ExceptionMiddleware
from faststream.kafka import KafkaBroker
from faststream.kafka.fastapi import KafkaRouter
from faststream.security import BaseSecurity, SASLPlaintext

from fast_clean.settings import CoreKafkaSettingsSchema
from fast_clean.utils import CertificateSchema, make_ssl_context


class BrokerFactory:
    """
    Фабрика брокеров сообщений.
    """

    router: KafkaRouter | None = None

    @classmethod
    def get_router(cls, kafka_settings: CoreKafkaSettingsSchema) -> KafkaRouter:
        if cls.router is None:
            cls.router = cls.make_router(kafka_settings)
        return cls.router

    @classmethod
    def make_static(cls, kafka_settings: CoreKafkaSettingsSchema) -> KafkaBroker:
        """
        Создаем брокер сообщений с помощью статического роутера.
        """
        return cls.get_router(kafka_settings).broker

    @classmethod
    @asynccontextmanager
    async def make_dynamic(cls, kafka_settings: CoreKafkaSettingsSchema) -> AsyncIterator[KafkaBroker]:
        """
        Создаем брокер сообщений с помощью динамического роутера.
        """
        async with cls.make_router(kafka_settings).broker as broker:
            yield broker

    @classmethod
    def make_router(cls, kafka_settings: CoreKafkaSettingsSchema) -> KafkaRouter:
        """
        Создаем роутер.
        """
        credentials = BrokerCredentials(kafka_settings)
        return KafkaRouter(
            kafka_settings.bootstrap_servers,
            security=credentials.get(),
            asyncapi_url='/asyncapi',
            middlewares=[cls.make_exception_middleware()],
        )

    @classmethod
    def make_exception_middleware(cls) -> ExceptionMiddleware:
        """
        Создаем middleware для обработки исключений.
        """
        exception_middleware = ExceptionMiddleware()
        exception_middleware.add_handler(Exception)(cls.exception_handler)
        return exception_middleware

    @staticmethod
    def exception_handler(exception: Exception) -> None:
        """
        Обработчик исключения.
        """
        print(repr(exception))


class BrokerCredentials:
    """
    Класс получения параметров для подключения к Kafka.
    """

    def __init__(self, kafka_settings: CoreKafkaSettingsSchema) -> None:
        self.kafka_settings = kafka_settings

    def get(self: Self, use_ssl: bool = True) -> BaseSecurity | SASLPlaintext | None:
        """
        Получаем параметры для подключения.
        """
        if self.kafka_settings.credentials is None:
            return self.get_none_credentials()

        credentials_mapping: dict[str, Callable[[], BaseSecurity | SASLPlaintext | None]] = {
            'SSL': partial(self.get_ssl_credentials, use_ssl),
            'SASL': partial(self.get_sasl_credentials, use_ssl),
        }
        return credentials_mapping.get(self.kafka_settings.credentials, self.get_none_credentials)()

    @staticmethod
    def get_none_credentials() -> None:
        """
        Получаем отсутствующие параметры для подключения.
        """
        return None

    def get_ssl_credentials(self: Self, use_ssl: bool = True) -> BaseSecurity:
        """
        Получаем SSL параметры для подключения.
        """
        ssl_context = make_ssl_context(CertificateSchema.model_validate(self.kafka_settings.model_dump()))
        return BaseSecurity(ssl_context, use_ssl)

    def get_sasl_credentials(self: Self, use_ssl: bool = True) -> SASLPlaintext:
        """
        Получаем SASL параметры для подключения.
        """
        assert self.kafka_settings.broker_username
        assert self.kafka_settings.broker_password
        return SASLPlaintext(
            username=self.kafka_settings.broker_username,
            password=self.kafka_settings.broker_password,
            use_ssl=use_ssl,
        )



"""
Пакет, содержащий репозиторий настроек.

Представлено 2 реализации:
- Env
- Prefect
"""

from typing import Protocol, Self

from .enums import SettingsSourceEnum
from .env import EnvSettingsRepository
from .exceptions import SettingsRepositoryError as SettingsRepositoryError
from .type_vars import SettingsSchema


class SettingsRepositoryProtocol(Protocol):
    """
    Протокол репозитория настроек.
    """

    async def get(self: Self, schema_type: type[SettingsSchema], *, name: str | None = None) -> SettingsSchema:
        """
        Получаем настройки.
        """
        ...


class SettingsRepositoryFactoryProtocol(Protocol):
    """
    Протокол фабрики репозиториев настроек.
    """

    async def make(self: Self, settings_source: SettingsSourceEnum) -> SettingsRepositoryProtocol:
        """
        Создаем репозиторий настроек.
        """
        ...


class SettingsRepositoryFactoryImpl:
    """
    Реализация фабрики репозиториев настроек.
    """

    async def make(self: Self, settings_source: SettingsSourceEnum) -> SettingsRepositoryProtocol:
        """
        Создаем репозиторий настроек.
        """
        match settings_source:
            case SettingsSourceEnum.ENV:
                return EnvSettingsRepository()



"""
Модуль, содержащий исключения.
"""

import traceback
import uuid
from abc import ABC, abstractmethod
from collections.abc import Iterable, Sequence
from functools import partial
from typing import Self, TypeVar

from fastapi import FastAPI, HTTPException, Request, Response, status
from fastapi.exception_handlers import http_exception_handler
from stringcase import camelcase, snakecase

from .enums import ModelActionEnum
from .schemas import BusinessLogicExceptionSchema, ModelAlreadyExistsErrorSchema, ValidationErrorSchema
from .settings import CoreSettingsSchema

ModelType = TypeVar('ModelType')


class ContainerError(Exception):
    """
    Ошибка контейнера зависимостей.
    """

    def __init__(self, message: str, *args: object) -> None:
        super().__init__(*args)
        self.message = message


class LockError(Exception):
    """
    Ошибка распределенной блокировки.
    """

    message = 'Errors acquiring or releasing a lock'


class BusinessLogicException(Exception, ABC):
    """
    Базовое исключение бизнес-логики.
    """

    @property
    def type(self: Self) -> str:
        """
        Тип ошибки.
        """
        return snakecase(type(self).__name__.replace('Error', ''))

    @property
    @abstractmethod
    def msg(self: Self) -> str:
        """
        Сообщение ошибки.
        """
        ...

    def __str__(self: Self) -> str:
        return self.msg

    def get_schema(self: Self, debug: bool) -> BusinessLogicExceptionSchema:
        """
        Получаем схему исключения.
        """
        return BusinessLogicExceptionSchema(
            type=self.type,
            msg=self.msg,
            traceback=(''.join(traceback.format_exception(type(self), self, self.__traceback__)) if debug else None),
        )


class PermissionDeniedError(BusinessLogicException):
    """
    Ошибка, возникающая при недостатке прав для выполнения действия.
    """

    @property
    def msg(self: Self) -> str:
        return 'Недостаточно прав для выполнения действия'


class ModelNotFoundError(BusinessLogicException):
    """
    Ошибка, возникающая при невозможности найти модель.
    """

    def __init__(
        self,
        model: type[ModelType] | str,
        *args: object,
        model_id: int | uuid.UUID | Iterable[int | uuid.UUID] | None = None,
        model_name: str | Iterable[str] | None = None,
        message: str | None = None,
    ) -> None:
        super().__init__(*args)
        self.model = model
        self.model_id = model_id
        self.model_name = model_name
        self.custom_message = message

    @property
    def msg(self: Self) -> str:
        if self.custom_message is not None:
            return self.custom_message
        msg = f'Не удалось найти модель {self.model if isinstance(self.model, str) else self.model.__name__}'
        if self.model_id is not None:
            if isinstance(self.model_id, Iterable):
                return f'{msg} по идентификаторам: [{", ".join(map(str, self.model_id))}]'
            return f'{msg} по идентификатору: {self.model_id}'
        if self.model_name is not None:
            if isinstance(self.model_name, Iterable):
                return f'{msg} по именам: [{", ".join(self.model_name)}]'
            return f'{msg} по имени: {self.model_name}'
        return msg


class ModelAlreadyExistsError(BusinessLogicException):
    """
    Ошибка, возникающая при попытке создать модель с существующим уникальным полем.
    """

    def __init__(self, field: str, message: str, *args: object) -> None:
        super().__init__(*args)
        self.field = field
        self.message = message

    @property
    def msg(self: Self) -> str:
        return self.message

    def get_schema(self: Self, debug: bool) -> BusinessLogicExceptionSchema:
        return ModelAlreadyExistsErrorSchema.model_validate(
            {**super().get_schema(debug).model_dump(), 'field': self.field}
        )


class ModelIntegrityError(BusinessLogicException):
    """
    Ошибка целостности данных при взаимодействии с моделью.
    """

    def __init__(self, model: type[ModelType] | str, action: ModelActionEnum, *args: object) -> None:
        super().__init__(*args)
        self.model = model
        self.action = action

    @property
    def msg(self: Self) -> str:
        """
        Сообщение ошибки.
        """
        msg = 'Ошибка целостности данных'
        model_name = self.model if isinstance(self.model, str) else self.model.__name__
        match self.action:
            case ModelActionEnum.INSERT:
                msg += f' при создании модели {model_name}'
            case ModelActionEnum.UPDATE:
                msg += f' при изменении модели {model_name}'
            case ModelActionEnum.UPSERT:
                msg += f' при создании или изменении модели {model_name}'
            case ModelActionEnum.DELETE:
                msg += f' при удалении модели {model_name}'
        return msg


class ValidationError(BusinessLogicException):
    """
    Ошибка валидации.
    """

    def __init__(self, field: str | Sequence[str], message: str, *args: object) -> None:
        super().__init__(*args)
        self.field = field
        self.message = message

    @property
    def fields(self: Self) -> Sequence[str]:
        """
        Поля ошибки.
        """
        return [self.field] if isinstance(self.field, str) else self.field

    @property
    def msg(self: Self) -> str:
        return self.message

    def get_schema(self: Self, debug: bool) -> BusinessLogicExceptionSchema:
        return ValidationErrorSchema.model_validate(
            {
                **super().get_schema(debug).model_dump(),
                'fields': list(map(camelcase, self.fields)),
            }
        )


class SortingFieldNotFoundError(BusinessLogicException):
    """
    Ошибка, возникающая при невозможности найти поле для сортировки.
    """

    def __init__(self, field: str, *args: object) -> None:
        super().__init__(*args)
        self.field = field

    @property
    def msg(self: Self) -> str:
        return f'Не удалось найти поле для сортировки: {self.field}'


async def business_logic_exception_handler(
    settings: CoreSettingsSchema, request: Request, exception: BusinessLogicException
) -> Response:
    """
    Обработчик базового исключения бизнес-логики.
    """
    return await http_exception_handler(
        request,
        HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail=[exception.get_schema(settings.debug).model_dump()],
        ),
    )


async def permission_denied_error_handler(
    settings: CoreSettingsSchema, request: Request, error: PermissionDeniedError
) -> Response:
    """
    Обработчик ошибки, возникающей при недостатке прав для выполнения действия.
    """
    return await http_exception_handler(
        request,
        HTTPException(
            status_code=status.HTTP_403_FORBIDDEN,
            detail=[error.get_schema(settings.debug).model_dump()],
        ),
    )


async def model_not_found_error_handler(
    settings: CoreSettingsSchema, request: Request, error: ModelNotFoundError
) -> Response:
    """
    Обработчик ошибки, возникающей при невозможности найти модель.
    """
    return await http_exception_handler(
        request,
        HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=[error.get_schema(settings.debug).model_dump()],
        ),
    )


async def model_already_exists_error_handler(
    settings: CoreSettingsSchema, request: Request, error: ModelAlreadyExistsError
) -> Response:
    """
    Обработчик ошибки, возникающей при попытке создать модель с существующим уникальным
    полем.
    """
    return await http_exception_handler(
        request,
        HTTPException(
            status_code=status.HTTP_409_CONFLICT,
            detail=[error.get_schema(settings.debug).model_dump()],
        ),
    )


def use_exceptions_handlers(app: FastAPI, settings: CoreSettingsSchema) -> None:
    """
    Регистрируем глобальные обработчики исключений.
    """
    app.exception_handler(BusinessLogicException)(partial(business_logic_exception_handler, settings))
    app.exception_handler(PermissionDeniedError)(partial(permission_denied_error_handler, settings))
    app.exception_handler(ModelNotFoundError)(partial(model_not_found_error_handler, settings))
    app.exception_handler(ModelAlreadyExistsError)(partial(model_already_exists_error_handler, settings))



"""
Модуль, содержащий репозиторий хранилища S3.
"""

import io
from collections.abc import AsyncGenerator
from contextlib import asynccontextmanager
from pathlib import Path
from typing import Self

import aiohttp
import miniopy_async
from miniopy_async.error import S3Error

from .reader import AiohttpStreamReader, StreamReaderProtocol, StreamReadProtocol
from .schemas import S3StorageParamsSchema


class S3StorageRepository:
    """
    Репозиторий хранилища S3.
    """

    def __init__(self: Self, params: S3StorageParamsSchema):
        self.params = params
        self.bucket = self.params.bucket
        self.client = miniopy_async.Minio(  # type: ignore
            f'{self.params.endpoint}:{self.params.port}',
            access_key=self.params.access_key,
            secret_key=self.params.secret_key,
            secure=self.params.secure,
        )

    async def exists(self: Self, path: str | Path) -> bool:
        """
        Проверяем существует ли файл.
        """
        try:
            await self.client.stat_object(self.bucket, self.get_str_path(path))
            return True
        except S3Error:
            return False

    async def listdir(self: Self, path: str | Path) -> list[str]:
        """
        Получаем список файлов и директорий в заданной директории.
        """
        str_path = self.get_str_path(path)
        if not str_path or str_path[-1] != '/':
            str_path += '/'
        objects = await self.client.list_objects(self.bucket, prefix=str_path if str_path != '/' else None)
        return [str(obj.object_name) for obj in objects] if objects else []

    async def is_file(self: Self, path: str | Path) -> bool:
        """
        Проверяем находится ли файл по пути.
        """
        return not await self.is_dir(path)

    async def is_dir(self: Self, path: str | Path) -> bool:
        """
        Проверяем находится ли директория по пути.
        """
        return len(await self.listdir(path)) > 0

    async def read(self: Self, path: str | Path) -> bytes:
        """
        Читаем содержимое файла.
        """
        async with aiohttp.ClientSession() as session:
            response: aiohttp.ClientResponse = await self.client.get_object(
                self.bucket, self.get_str_path(path), session
            )
            return await response.read()

    @asynccontextmanager
    async def stream_read(self: Self, path: str | Path) -> AsyncGenerator[StreamReaderProtocol, None]:
        """
        Читаем содержимое файла в потоковом режиме.
        """
        async with aiohttp.ClientSession() as session:
            reader = await self.client.get_object(self.bucket, self.get_str_path(path), session)
            yield AiohttpStreamReader(reader)

    async def write(self: Self, path: str | Path, content: str | bytes) -> None:
        """
        Создаем файл или переписываем существующий.
        """
        content = content.encode('utf-8') if isinstance(content, str) else content
        data = io.BytesIO(content)
        await self.client.put_object(self.bucket, self.get_str_path(path), data, len(content))

    async def stream_write(
        self: Self,
        path: str | Path,
        stream: StreamReadProtocol,
        length: int = -1,
        part_size: int = 0,
    ) -> None:
        """
        Создаем файл или переписываем существующий в потоковом режиме.
        """
        await self.client.put_object(self.bucket, self.get_str_path(path), stream, length=length, part_size=part_size)

    async def delete(self: Self, path: str | Path) -> None:
        """
        Удаляем файл.
        """
        await self.client.remove_object(self.bucket, self.get_str_path(path))

    @staticmethod
    def get_str_path(path: str | Path) -> str:
        """
        Получаем путь в виде строки.
        """
        if isinstance(path, Path):
            return '/' if path == Path('') else str(path)
        return path



"""
Модуль, содержащий настройки.
"""

from __future__ import annotations

import os
from pathlib import Path
from typing import ClassVar, Literal, Self

from pydantic import BaseModel, ConfigDict, RedisDsn, model_validator
from pydantic_settings import BaseSettings as PydanticBaseSettings
from pydantic_settings import SettingsConfigDict
from typing_extensions import Unpack


class CoreDbSettingsSchema(BaseModel):
    """
    Схема настроек базы данных.
    """

    provider: str = 'postgresql+psycopg_async'

    host: str
    port: int
    user: str
    password: str
    name: str

    scheme: str = 'public'

    @property
    def dsn(self: Self) -> str:
        """
        DSN подключения к базе данных.
        """
        return f'{self.provider}://{self.user}:{self.password}@{self.host}:{self.port}/{self.name}'


class CoreCacheSettingsSchema(BaseModel):
    """
    Схема настроек кеша.
    """

    provider: Literal['in_memory', 'redis'] = 'in_memory'

    prefix: str


class CoreS3SettingsSchema(BaseModel):
    """
    Схема настроек S3.
    """

    endpoint: str
    endpoint_url: str
    access_key: str
    secret_key: str
    port: int
    bucket: str
    secure: bool = False


class CoreStorageSettingsSchema(BaseModel):
    """
    Схема настроек хранилища.
    """

    provider: Literal['local', 's3'] = 'local'

    dir: Path = Path(__file__).resolve().parent.parent / 'storage'
    s3: CoreS3SettingsSchema | None = None


class CoreElasticsearchSettingsSchema(BaseModel):
    """
    Схема настроек Elasticsearch.
    """

    host: str
    port: int
    scheme: str
    username: str
    password: str
    cluster_name: str

    cacert: str | None = None
    security: bool = False
    ssl: bool = False


class CoreSearchSettingsSchema(BaseModel):
    """
    Схема настроек движка поиска.
    """

    provider: Literal['elasticsearch', 'open_search'] = 'elasticsearch'
    elasticsearch: CoreElasticsearchSettingsSchema | None = None


class CoreKafkaSettingsSchema(BaseModel):
    """
    Схема настроек Kafka.
    """

    bootstrap_servers: str
    group_id: str
    credentials: Literal['SSL', 'SASL'] | None = None
    # For SSL
    cert_file: str | None = None
    ca_file: str | None = None
    key_file: str | None = None
    password: str | None = None
    # For SASL
    broker_username: str | None = None
    broker_password: str | None = None

    @model_validator(mode='after')
    def validate_credentials(self: Self) -> Self:
        """
        Проверяем на правильность заполнение параметров для авторизации.
        """
        match self.credentials:
            case 'SSL':
                for field in ('ca_file', 'cert_file', 'key_file'):
                    assert bool(getattr(self, field)), f'{field} must be set when credentials={self.credentials}'
            case 'SASL':
                for field in ('broker_username', 'broker_password'):
                    assert bool(getattr(self, field)), f'{field} must be set when credentials={self.credentials}'
            case _:
                ...
        return self


class CoreServiceSettingsSchema(BaseModel):
    """
    Схема настроек доступа к сервису.
    """

    host: str
    user: str | None = None
    password: str | None = None


class CoreTopicSettingsSchema(BaseModel):
    """
    Схема настроек топика.
    """

    name: str
    auto_offset_reset: Literal['latest', 'earliest', 'none'] = 'latest'


class BaseSettingsSchema(PydanticBaseSettings):
    """
    Схема настроек с возможностью поиска через репозиторий.
    """

    descendant_types: ClassVar[list[type[BaseSettingsSchema]]] = []

    def __init_subclass__(cls, **kwargs: Unpack[ConfigDict]) -> None:
        """
        Добавляем настройки в репозиторий.
        """
        cls.descendant_types.append(cls)

        return super().__init_subclass__(**kwargs)


class CoreSettingsSchema(BaseSettingsSchema):
    """
    Схема базовых настроек приложения.
    """

    debug: bool
    title: str
    base_url: str
    base_dir: Path = Path(os.getcwd())
    secret_key: str
    cors_origins: list[str]
    redis_dsn: RedisDsn | None = None

    model_config = SettingsConfigDict(
        env_file='.env',
        env_file_encoding='utf-8',
        env_nested_delimiter='__',
        case_sensitive=False,
        extra='ignore',
    )



"""
Модуль, содержащий переменные типов.
"""

import uuid
from typing import TypeVar

from fast_clean.db import BaseInt, BaseUUID
from fast_clean.schemas import (
    CreateSchema,
    CreateSchemaInt,
    ReadSchema,
    ReadSchemaInt,
    UpdateSchema,
    UpdateSchemaInt,
)

ModelBaseType = TypeVar('ModelBaseType', bound=BaseInt | BaseUUID)
CreateSchemaBaseType = TypeVar('CreateSchemaBaseType', bound=CreateSchemaInt | CreateSchema)
ReadSchemaBaseType = TypeVar('ReadSchemaBaseType', bound=ReadSchemaInt | ReadSchema)
UpdateSchemaBaseType = TypeVar('UpdateSchemaBaseType', bound=UpdateSchemaInt | UpdateSchema)
IdType = TypeVar('IdType', bound=int | uuid.UUID)
IdTypeContravariant = TypeVar('IdTypeContravariant', bound=int | uuid.UUID, contravariant=True)


ModelIntType = TypeVar('ModelIntType', bound=BaseInt)
CreateSchemaIntType = TypeVar('CreateSchemaIntType', bound=CreateSchemaInt)
ReadSchemaIntType = TypeVar('ReadSchemaIntType', bound=ReadSchemaInt)
UpdateSchemaIntType = TypeVar('UpdateSchemaIntType', bound=UpdateSchemaInt)


ModelType = TypeVar('ModelType', bound=BaseUUID)
CreateSchemaType = TypeVar('CreateSchemaType', bound=CreateSchema)
ReadSchemaType = TypeVar('ReadSchemaType', bound=ReadSchema)
UpdateSchemaType = TypeVar('UpdateSchemaType', bound=UpdateSchema)



"""
Модуль, содержащий зависимости.
"""

import json
from collections.abc import AsyncIterator, Sequence
from typing import Annotated

from dishka import Provider, Scope, provide
from fastapi import Depends, Request
from faststream.kafka import KafkaBroker
from flatten_dict import unflatten
from sqlalchemy.ext.asyncio import AsyncSession
from starlette.datastructures import FormData
from stringcase import snakecase

from .broker import BrokerFactory
from .db import SessionFactory, SessionManagerImpl, SessionManagerProtocol
from .redis import RedisManager
from .repositories import (
    CacheManager,
    CacheRepositoryProtocol,
    LocalStorageParamsSchema,
    S3StorageParamsSchema,
    SettingsRepositoryFactoryImpl,
    SettingsRepositoryFactoryProtocol,
    SettingsRepositoryProtocol,
    SettingsSourceEnum,
    StorageRepositoryFactoryImpl,
    StorageRepositoryFactoryProtocol,
    StorageRepositoryProtocol,
    StorageTypeEnum,
)
from .schemas import PaginationRequestSchema
from .services import (
    CryptographicAlgorithmEnum,
    CryptographyServiceFactory,
    CryptographyServiceProtocol,
    LockServiceProtocol,
    RedisLockService,
    SeedService,
    TransactionService,
)
from .settings import CoreCacheSettingsSchema, CoreKafkaSettingsSchema, CoreSettingsSchema, CoreStorageSettingsSchema


async def get_nested_form_data(request: Request) -> FormData:
    """
    Получаем форму, позволяющую использовать вложенные словари.
    """
    dot_data = {k.replace('[', '.').replace(']', ''): v for k, v in (await request.form()).items()}
    nested_data = unflatten(dot_data, 'dot')
    for k, v in nested_data.items():
        if isinstance(v, dict):
            nested_data[k] = json.dumps(v)
    return FormData(nested_data)


def get_pagination(page: int | None = None, page_size: int | None = None) -> PaginationRequestSchema:
    """
    Получаем входные данные пагинации.
    """
    return PaginationRequestSchema(page=page or 1, page_size=page_size or 10)


def get_sorting(sorting: str | None = None) -> Sequence[str]:
    """
    Получаем входные данные сортировки.
    """
    if not sorting:
        return []
    return [s[0] + snakecase(s[1:]) if s[0] == '-' else snakecase(s) for s in sorting.split(',')]


NestedFormData = Annotated[FormData, Depends(get_nested_form_data)]
Pagination = Annotated[PaginationRequestSchema, Depends(get_pagination)]
Sorting = Annotated[Sequence[str], Depends(get_sorting)]


class CoreProvider(Provider):
    """
    Провайдер зависимостей.
    """

    scope = Scope.REQUEST

    # --- repositories ---

    settings_repository_factory = provide(
        SettingsRepositoryFactoryImpl, provides=SettingsRepositoryFactoryProtocol, scope=Scope.APP
    )
    storage_repository_factory = provide(
        StorageRepositoryFactoryImpl, provides=StorageRepositoryFactoryProtocol, scope=Scope.APP
    )

    @provide(scope=Scope.APP)
    @staticmethod
    async def get_settings_repository(
        settings_repository_factory: SettingsRepositoryFactoryProtocol,
    ) -> SettingsRepositoryProtocol:
        """
        Получаем репозиторий настроек.
        """
        return await settings_repository_factory.make(SettingsSourceEnum.ENV)

    @provide(scope=Scope.APP)
    @staticmethod
    async def get_settings(settings_repository: SettingsRepositoryProtocol) -> CoreSettingsSchema:
        """
        Получаем настройки.
        """
        return await settings_repository.get(CoreSettingsSchema)

    @staticmethod
    async def get_broker_repository(settings_repository: SettingsRepositoryProtocol) -> AsyncIterator[KafkaBroker]:
        """
        Получаем репозиторий брокера сообщений.
        """
        kafka_settings = await settings_repository.get(CoreKafkaSettingsSchema)
        yield BrokerFactory.make_static(kafka_settings)

    @provide(scope=Scope.APP)
    @staticmethod
    async def get_cache_repository(settings_repository: SettingsRepositoryProtocol) -> CacheRepositoryProtocol:
        """
        Получаем репозиторий кеша.
        """
        settings = await settings_repository.get(CoreSettingsSchema)
        if settings.redis_dsn is not None:
            RedisManager.init(settings.redis_dsn)
        cache_settings = await settings_repository.get(CoreCacheSettingsSchema)
        if CacheManager.cache_repository is None:
            CacheManager.init(cache_settings, RedisManager.redis)
        if CacheManager.cache_repository is not None:
            return CacheManager.cache_repository
        raise ValueError('Cache is not initialized')

    @provide(scope=Scope.APP)
    @staticmethod
    async def get_storage_repository(
        settings_repository: SettingsRepositoryProtocol,
        storage_repository_factory: StorageRepositoryFactoryProtocol,
    ) -> StorageRepositoryProtocol:
        """
        Получаем репозиторий файлового хранилища.
        """
        storage_settings = await settings_repository.get(CoreStorageSettingsSchema)
        if storage_settings.provider == 's3' and storage_settings.s3 is not None:
            return await storage_repository_factory.make(
                StorageTypeEnum.S3,
                S3StorageParamsSchema.model_validate(storage_settings.s3.model_dump()),
            )
        elif storage_settings.provider == 'local' and storage_settings.dir is not None:
            return await storage_repository_factory.make(
                StorageTypeEnum.LOCAL, LocalStorageParamsSchema(path=storage_settings.dir)
            )
        raise NotImplementedError(f'Storage {storage_settings.provider} not allowed')

    # --- db ---

    @provide
    @staticmethod
    async def get_async_session(settings_repository: SettingsRepositoryProtocol) -> AsyncIterator[AsyncSession]:
        """
        Получаем асинхронную сессию.
        """
        async with SessionFactory.make_async_session_static(settings_repository) as session:
            yield session

    @provide
    @staticmethod
    def get_session_manager(session: AsyncSession) -> SessionManagerProtocol:
        """
        Получаем менеджер сессий.
        """
        return SessionManagerImpl(session)

    # --- services ---

    seed_service = provide(SeedService)
    transaction_service = provide(TransactionService)

    @provide(scope=Scope.APP)
    @staticmethod
    def get_cryptography_service_factory(settings: CoreSettingsSchema) -> CryptographyServiceFactory:
        """
        Получаем фабрику сервисов криптографии.
        """
        return CryptographyServiceFactory(settings.secret_key)

    @provide(scope=Scope.APP)
    @staticmethod
    async def get_cryptography_service(
        cryptography_service_factory: CryptographyServiceFactory,
    ) -> CryptographyServiceProtocol:
        """
        Получаем сервис криптографии.
        """
        return await cryptography_service_factory.make(CryptographicAlgorithmEnum.AES_GCM)

    @provide(scope=Scope.APP)
    @staticmethod
    def get_lock_service(settings: CoreSettingsSchema) -> LockServiceProtocol:
        """
        Получаем сервис распределенной блокировки.
        """
        assert settings.redis_dsn is not None
        RedisManager.init(settings.redis_dsn)
        assert RedisManager.redis is not None
        return RedisLockService(RedisManager.redis)


provider = CoreProvider()



"""
Модуль, содержащий перечисления.
"""

from enum import StrEnum, auto
from typing import Self


class CascadeEnum(StrEnum):
    """
    Настройки каскадного поведения для SQLAlchemy.
    """

    SAVE_UPDATE = 'save-update'
    MERGE = 'merge'
    REFRESH_EXPIRE = 'refresh-expire'
    EXPUNGE = 'expunge'
    DELETE = 'delete'
    ALL = 'all'
    DELETE_ORPHAN = 'delete-orphan'

    ALL_DELETE_ORPHAN = 'all, delete-orphan'

    def __add__(self: Self, value: str) -> str:
        return f'{self}, {value}'

    def __radd__(self: Self, value: str) -> str:
        return f'{value}, {self}'


class ModelActionEnum(StrEnum):
    """
    Действие с моделью.
    """

    INSERT = auto()
    UPDATE = auto()
    UPSERT = auto()
    DELETE = auto()



"""
Модуль, содержащий контейнер зависимостей.
"""

import importlib
import os
import sys
from collections.abc import AsyncIterator
from contextlib import asynccontextmanager
from pathlib import Path

from dishka import AsyncContainer, Provider, make_async_container
from dishka.integrations.fastapi import FastapiProvider, setup_dishka
from fastapi import FastAPI


class ContainerManager:
    """
    Менеджер для управления контейнером зависимостей.
    """

    DEPENDS_MODULE = 'depends'

    container: AsyncContainer | None = None

    @classmethod
    def init(cls, module_names: set[str] | None = None) -> AsyncContainer:
        """
        Инициализируем контейнер зависимостей.
        """
        if cls.container is None:
            cls.container = cls.create(module_names)
        return cls.container

    @classmethod
    def init_for_fastapi(cls, app: FastAPI, module_names: set[str] | None = None) -> AsyncContainer:
        """
        Инициализируем контейнер зависимостей для приложения FastAPI.
        """
        container = cls.init(module_names)
        setup_dishka(container=container, app=app)
        return container

    @classmethod
    async def close(cls) -> None:
        """
        Закрываем контейнер зависимостей.
        """
        if cls.container is None:
            return
        await cls.container.close()
        cls.container = None

    @classmethod
    def create(cls, module_names: set[str] | None = None) -> AsyncContainer:
        """
        Создаем контейнер зависимостей.
        """
        module_names = module_names or set()
        module_names.update(cls.get_default_module_names())
        return make_async_container(FastapiProvider(), *cls.get_providers(module_names))

    @classmethod
    def get_default_module_names(cls) -> set[str]:
        """
        Получаем список модулей с зависимостями по умолчанию.
        """
        cwd = Path(os.getcwd())
        virtual_env_paths = {path.parent for path in cwd.rglob('pyvenv.cfg')}
        module_names: set[str] = set()
        for path in cwd.rglob(f'{cls.DEPENDS_MODULE}.py'):
            if not any(path.is_relative_to(venv) for venv in virtual_env_paths):
                module_names.add('.'.join(str(path.relative_to(cwd).with_suffix('')).split('/')))
        module_names.add(f'fast_clean.{cls.DEPENDS_MODULE}')
        return module_names

    @staticmethod
    def get_providers(module_names: set[str]) -> list[Provider]:
        """
        Получаем провайдеры зависимостей.
        """
        providers: list[Provider] = []
        for module_name in module_names:
            module = sys.modules[module_name] if module_name in sys.modules else importlib.import_module(module_name)
            for obj in module.__dict__.values():
                if isinstance(obj, Provider):
                    providers.append(obj)
        return providers


@asynccontextmanager
async def get_container() -> AsyncIterator[AsyncContainer]:
    """
    Получаем контейнер зависимостей.
    """
    container = ContainerManager.container
    if container is None:
        container = ContainerManager.init()
    async with container() as nested_container:
        yield nested_container



from typing import Any

import sqlalchemy_utils
from alembic.autogenerate.api import AutogenContext


def render_item(type_: str, obj: Any, autogen_context: AutogenContext):
    """
    Apply custom rendering for selected items.
    """

    if type_ == 'type' and isinstance(obj, sqlalchemy_utils.types.ChoiceType):
        autogen_context.imports.add('import sqlalchemy_utils')
        autogen_context.imports.add(f'from {obj.choices.__module__} import {obj.choices.__name__}')
        return f'sqlalchemy_utils.types.ChoiceType({obj.choices.__name__}, impl=sa.{obj.impl.__class__.__name__}())'

    if type_ == 'type' and isinstance(obj, sqlalchemy_utils.types.JSONType):
        autogen_context.imports.add('import sqlalchemy_utils')
        return 'sqlalchemy_utils.types.JSONType()'

    if type_ == 'type' and isinstance(obj, sqlalchemy_utils.types.UUIDType):
        autogen_context.imports.add('import sqlalchemy_utils')
        return f'sqlalchemy_utils.types.UUIDType(binary={obj.binary})'

    # Default rendering for other objects
    return False



"""
Модель запуска тяжелых операций в ThreadPoolExecutor.
"""

import asyncio
from functools import partial
from typing import Callable, TypeVar

from typing_extensions import ParamSpec

R = TypeVar('R')
P = ParamSpec('P')


async def run_in_threadpool(fn: Callable[P, R], *args: P.args, **kwargs: P.kwargs) -> R:
    """
    Запуск функции в отдельном потоке.
    """
    kwargs_fn = partial(fn, *args, **kwargs)
    loop = asyncio.get_running_loop()
    return await loop.run_in_executor(None, kwargs_fn)
